﻿using System.Collections.Generic;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.EngineCore;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class MapObjectLayer manages the display of a list of MapObjects as a layer in ArcGIS
    /// </summary>
    class MapObjectLayer
    {
        /** The name of the layer */
        private string name;
        /** The host of the 3D Mapping viewer */
        private PluginArcMap viewerHost;

        /** The map we are displaying this layer on */
        private ESRI.ArcGIS.Carto.IMap map;
        /** The layer with overlay graphics */
        private IGraphicTracker graphicTracker;
        /** The current map object overlays */
        private List<MapObjectOverlay> mapOverlays;
        /** The new map object overlays waiting to be displayed */
        private AList<MapObject> newMapObjects;
        /** Visible? */
        private bool visible;

        /** The last extent of the map display */
        private double viewX0,viewY0,viewX1,viewY1;

        /// <summary>
        /// Create a new layer
        /// </summary>
        /// <param name="name">the name of the layer</param>
        /// <param name="viewerHost">the host of the 3D Mapping viewer</param>
        internal MapObjectLayer(string name,PluginArcMap viewerHost)
        {
            this.name = name;
            this.viewerHost = viewerHost;
            this.mapOverlays = new List<MapObjectOverlay>();
            this.newMapObjects = null;
            this.visible = true;
            /* Create the graphics tracker */
            this.map = viewerHost.getMap();
            this.graphicTracker = new GraphicTrackerClass();
            this.graphicTracker.Initialize(this.map);
            /* Request a map update */
            this.viewX0 = 0.0;
            this.viewY0 = 0.0;
            this.viewX1 = -1.0;
            this.viewX1 = -1.0;
            this.viewerHost.RequestMapUpdate("added mapobject layer '"+this.name+"'");
        }

        /// <summary>
        /// Get the graphics tracker
        /// </summary>
        /// <returns></returns>
        internal IGraphicTracker getGraphics()
        {
            return this.graphicTracker;
        }

        /// <summary>
        /// Get the map that displays the graphics.
        /// </summary>
        /// <returns></returns>
        internal ESRI.ArcGIS.Carto.IMap getMap()
        {
            return this.map;
        }

        /// <summary>
        /// Change the visibility of the layer
        /// </summary>
        /// <param name="visible"></param>
        internal void setVisible(bool visible)
        {
            this.visible = visible;
            /* Start the updates */
            this.graphicTracker.SuspendUpdate = true;
            /* Draw/update the overlays */
            List<MapObjectOverlay> currentOverlays = this.mapOverlays;
            foreach (MapObjectOverlay mapOverlay in currentOverlays) mapOverlay.setVisible(visible);
            /* Stop the updates */
            this.graphicTracker.SuspendUpdate = false;
        }

        /// <summary>
        /// Update the layer to display a new list of map objects
        /// Calling this method does not trigger a map-view update, the caller should trigger one
        /// </summary>
        internal void update(AList<MapObject> mapObjects)
        {
            Message.WriteLine("updating mapobject layer '" + this.name + "'");
            if (mapObjects == null) return;
            this.newMapObjects = mapObjects;
        }

        /// <summary>
        /// Unload the layer
        /// </summary>
        internal void unload()
        {
            /* Just clear the list of map objects */
            this.update(new AList<MapObject>(MapObject.TYPE));
            /* Request a map update */
            this.viewerHost.RequestMapUpdate("unloaded mapobject layer '" + this.name + "'");
        }

        /// <summary>
        /// Draw the layer
        /// </summary>
        /// <param name="display">the display to overlay on</param>
        internal void Draw(IDisplay display)
        {
            /* Did the view change? */
            IEnvelope viewExtent = display.DisplayTransformation.VisibleBounds;
            bool movedView = false;
            if (viewExtent.XMin != this.viewX0) movedView = true;
            if (viewExtent.YMin != this.viewY0) movedView = true;
            if (viewExtent.XMax != this.viewX1) movedView = true;
            if (viewExtent.YMax != this.viewY1) movedView = true;
            this.viewX0 = viewExtent.XMin;
            this.viewY0 = viewExtent.YMin;
            this.viewX1 = viewExtent.XMax;
            this.viewY1 = viewExtent.YMax;
            /* Visible? */
            if (this.visible == false) return;
            /* Start the updates */
            this.graphicTracker.SuspendUpdate = true;
            /* Checkin new map objects? */
            AList<MapObject> newMapObjects2 = this.newMapObjects;
            this.newMapObjects = null;
            if (newMapObjects2 != null)
            {
                foreach (MapObjectOverlay mapObjectOverlay in this.mapOverlays) mapObjectOverlay.Close();
                List<MapObjectOverlay> newOverlays = new List<MapObjectOverlay>();
                foreach (MapObject mapObject in newMapObjects2.toArray()) newOverlays.Add(new MapObjectOverlay(this, mapObject));
                this.mapOverlays = newOverlays;
            }
            /* Draw/update the overlays */
            List<MapObjectOverlay> currentOverlays = this.mapOverlays;
            Message.WriteLine("DRAWING " + currentOverlays.Count + " overlays of mapobject layer '" + this.name + "', view moved? "+ movedView);
            foreach (MapObjectOverlay mapOverlay in currentOverlays) mapOverlay.draw(display,movedView);
            /* Stop the updates */
            this.graphicTracker.SuspendUpdate = false;
        }
    }
}
