﻿using System;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class Overlays manages the overlay of ArcGIS features on the 3D Mapping viewer
    /// </summary>
    class Overlays
    {
        // define the (outline) color of the overlays
        static private Color COLOR = Color.fromRGBA(0, 0, 255, 255);
        // define the fill color of the overlays
        static private Color FILL_COLOR = Color.fromRGBA(0, 0, 255, 128);
        // define the outline line width
        static private double LINE_WIDTH = 1.0;

        // Convert a point list to a part
        //
        static private CoordinatePart pointListToPart(string crs, IPointCollection pointList,bool hasZ)
        {
            CoordinatePart part = new CoordinatePart();
            part.crs = crs;
            for (int i=0; i< pointList.PointCount; i++)
            {
                IPoint point = pointList.Point[i];
                Coordinate coordinate;
                if (hasZ) coordinate = Coordinate.fromXYZ(crs, point.X, point.Y, point.Z);
                else coordinate = Coordinate.fromXY(crs, point.X, point.Y);
                part.points.add(coordinate);
            }
            return part;
        }

        // Create new map objects from a feature
        //
        static private void createMapObjects(IFeature feature,string layerName, AList<MapObject> mapObjects)
        {
            IGeometry geometry = feature.Shape;
            if (geometry==null)
            {
                Message.WriteLine("Feature has no geometry");
                return;
            }
            ISpatialReference layerSR = geometry.SpatialReference;
            if (layerSR == null)
            {
                Message.WriteLine("Feature has no spatial reference");
                return;
            }
            string crs = "" + layerSR.FactoryCode;
            bool hasZ = false;
            if (geometry is IZAware) hasZ = (geometry as IZAware).ZAware;
            String objectId = layerName + "#" + feature.OID;
            // do we have a polygon?
            if ((geometry.GeometryType==esriGeometryType.esriGeometryPolygon) && (geometry is IGeometryCollection))
            {
                IGeometryCollection partList = geometry as IGeometryCollection;
                Message.WriteLine("Found polygon (CRS " + crs + ")");
                CoordinatePath path = new CoordinatePath();
                path.crs = crs;
                for (int i = 0; i < partList.GeometryCount; i++)
                {
                    IPointCollection part = partList.Geometry[i] as IPointCollection;
                    CoordinatePart cpart = pointListToPart(crs, part, hasZ);
                    if (cpart != null) path.parts.add(cpart);
                }
                MapObject mapObject = new MapObject();
                mapObject.spatialArea = path;
                mapObject.color = COLOR;
                mapObject.fillColor = FILL_COLOR;
                mapObject.lineWidth = LINE_WIDTH;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            // do we have a polyline?
            if ((geometry.GeometryType == esriGeometryType.esriGeometryPolyline) && (geometry is IGeometryCollection))
            {
                IGeometryCollection partList = geometry as IGeometryCollection;
                Message.WriteLine("Found polyline (CRS " + crs + ")");
                CoordinatePath path = new CoordinatePath();
                path.crs = crs;
                for (int i = 0; i < partList.GeometryCount; i++)
                {
                    IPointCollection part = partList.Geometry[i] as IPointCollection;
                    CoordinatePart cpart = pointListToPart(crs, part, hasZ);
                    if (cpart != null) path.parts.add(cpart);
                }
                MapObject mapObject = new MapObject();
                mapObject.spatialLine = path;
                mapObject.color = COLOR;
                mapObject.lineWidth = LINE_WIDTH;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            // do we have a point?
            if ((geometry.GeometryType == esriGeometryType.esriGeometryPoint) && (geometry is IPoint))
            {
                IPoint point = (IPoint)geometry;
                Message.WriteLine("Found point (CRS "+ crs+")");
                MapObject mapObject = new MapObject();
                if (hasZ) mapObject.spatialPoint = Coordinate.fromXYZ(crs, point.X, point.Y, point.Z);
                else mapObject.spatialPoint = Coordinate.fromXY(crs, point.X, point.Y);
                mapObject.color = COLOR;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            // do we have a multi-point?
            if (geometry is Multipoint)
            {
                Multipoint mpoint = (Multipoint)geometry;
                Message.WriteLine("Found multipoint (CRS " + crs + ")");
                for (int i=0; i<mpoint.PointCount; i++)
                {
                    IPoint point = mpoint.Point[i];
                    MapObject mapObject = new MapObject();
                    if (hasZ) mapObject.spatialPoint = Coordinate.fromXYZ(crs, point.X, point.Y, point.Z);
                    else mapObject.spatialPoint = Coordinate.fromXY(crs, point.X, point.Y);
                    mapObject.color = COLOR;
                    mapObject.objectId = objectId + "#" + i;
                    mapObjects.add(mapObject);
                }
                return;
            }
            Message.WriteLine("Feature not added as overlay, the geometry class is: " + geometry.GetType().Name);
            return;
        }

        /// <summary>
        /// Add the selected ArcGIS features as overlays on the 3D Mapping viewer
        /// </summary>
        /// <param name="mapView"></param>
        /// <param name="viewer"></param>
        static internal void doAddOverlays(ESRI.ArcGIS.Carto.IMap mapView, Viewer3DM viewer)
        {
            try
            {
                // create a map object for each selected feature
                Message.WriteLine("Adding the selected features as overlays");
                AList<MapObject> mapObjects = new AList<MapObject>(MapObject.TYPE);
                IEnumFeature selection = mapView.FeatureSelection as IEnumFeature;
                selection.Reset();
                IFeature feature = selection.Next();
                while (feature != null)
                {
                    string layerName = (feature.Class as IDataset).Name;
                    Message.WriteLine("Adding selected feature "+feature.OID+" of '"+layerName+"'");
                    createMapObjects(feature, layerName, mapObjects);
                    feature = selection.Next();
                }
                Message.WriteLine("Created " + mapObjects.size() + " map objects");
                // send the map objects to the viewer
                if (mapObjects.size() > 0) viewer.addOverlayObjects(mapObjects);
                else viewer.clearOverlayObjects();
                // clear the feature selection to indicate we have processed the objects
                mapView.ClearSelection();
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while adding overlays: " + e);
            }
        }
    }
}
