﻿using System;
using Directory = System.IO.Directory;
using File = System.IO.File;
using FileInfo = System.IO.FileInfo;
using FileMode = System.IO.FileMode;
using FileStream = System.IO.FileStream;
using IOPath = System.IO.Path;
using StreamWriter = System.IO.StreamWriter;
using SynchronizationContext = System.Threading.SynchronizationContext;
using System.Timers;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;
using OrbitGT.Mapping3D.SDK;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class PluginArcMap implements the 3D Mapping Viewer plugin for ArcMap
    /// </summary>
    class PluginArcMap : SDKViewerApplication3DM
    {
        /** The single instance of this class */
        static internal PluginArcMap INSTANCE = null;

        /** The name of the optional configuration file */
        static private string CONFIG_FILENAME = "C:\\plugins\\plugin_arcmap.ini";

        private SDKViewerHost viewerHost;
        private MapObjectLayer footprintLayer;
        private MapObjectLayer fieldOfViewLayer;
        private MapObjectLayer measureLayer;
        private MapObjectLayer globalLayer;
        private UpdateFootprint footprintUpdater;
        private UpdateFieldOfView fieldOfViewUpdater;
        private UpdateMeasurement measurementUpdater;
        private UpdateGlobal globalUpdater;
        private Timer refreshTimer;

        // the main UI thread context
        private SynchronizationContext uiThread;
        // the map view that shows the viewer overlays
        private ESRI.ArcGIS.Carto.IMap map0;
        // the old view bounds to make refresh possible
        private double viewX0, viewY0, viewX1, viewY1;
        private IEnvelope viewBounds;
        private Bounds mapExtent;
        private double mapScale;

        /// <summary>
        /// Create a new plugin instance
        /// </summary>
        public PluginArcMap()
        {
            try
            {
                // create a logfile
                string userDir = IOPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
                if (Directory.Exists(userDir) == false) Directory.CreateDirectory(userDir);
                string logFileName = IOPath.Combine(userDir, "arcmap_3dmc_log.txt");
                Message.setLogFile(new StreamWriter(logFileName));
                Message.WriteLine("Creating ArcMap plugin");
                Message.WriteLine("Logging to file '" + logFileName + "'");
                Message.WriteLine("Start time " + DateTime.Now.ToString("yyyy.MM.dd HH:mm:ss"));
                Message.WriteLine("Machine name is '" + System.Environment.MachineName + "'");
                string assemblyCachePath = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
                Message.WriteLine("Assembly cache path is '" + assemblyCachePath + "'");
                // read the optional configuration file
                Message.WriteLine("Reading configuration file '" + CONFIG_FILENAME + "'");
                AMap<string> configuration = this.readConfigurationFile();
                foreach (string key in configuration.listKeys().toArray()) Message.WriteLine("Read configuration '" + key + "' value '" + configuration.get(key) + "'");
                // get the main UI thread context to call UI methods from background threads
                this.uiThread = SynchronizationContext.Current;
                // initialize the mapping system
                IMxDocument mxDocument = (IMxDocument)ArcMap.Application.Document;
                this.map0 = mxDocument.ActiveView.FocusMap;
                ISpatialReferenceFactory3 crsFactory = new SpatialReferenceEnvironmentClass();
                MapObjects.VIEWER_CRS = crsFactory.CreateSpatialReference(MapObjects.VIEWER_CRS_CODE);
                // initialize the 3D Mapping viewer (starts the viewer host and the websocket server)
                // we pass the auto-login application state (remember-be) and coordinate system preference using the startup options
                this.viewerHost = new SDKViewerHost();
                AMap<object> viewerOptions = new AMap<object>(Collections.TYPE_OBJECT);
                string appState = this.readApplicationStateFile();
                if (appState != null) viewerOptions.set(orbitgt.mapping3d.sdk.viewer.Constants.STARTUP_APP_STATE, appState);
                viewerOptions.set(orbitgt.mapping3d.sdk.viewer.Constants.STARTUP_CRS, MapObjects.CRS);
                foreach(string key in configuration.listKeys().toArray()) viewerOptions.set(key, configuration.get(key));
                Message.WriteLine("Initializing viewer");
                this.viewerHost.InitializeViewer(this, viewerOptions);
                Message.WriteLine("Initialized viewer");
                // listen to various signals emitted by the 3D Mapping viewer
                this.getViewer().onAppStateChanged.SignalEvent += OnAppStateChanged;
                this.getViewer().onFootprintsChanged.SignalEvent += OnFootprintsChanged;
                this.getViewer().onFieldOfViewsChanged.SignalEvent += OnFieldOfViewsChanged;
                this.getViewer().onMeasureStateChanged.SignalEvent += OnMeasureStateChanged;
                this.getViewer().onGlobalsChanged.SignalEvent += OnGlobalsChanged;
                this.getViewer().onZoomToExtent.SignalEvent += OnZoomToExtent;
                this.getViewer().onPickPositionRequested.SignalEvent += OnPickPositionRequested;
                this.getViewer().onPickPositionCancelled.SignalEvent += OnPickPositionCancelled;
                // start the viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
                Message.WriteLine("Starting viewer");
                this.viewerHost.StartViewer();
                Message.WriteLine("Started viewer");
                // create the layers
                Message.WriteLine("Creating map layers");
                this.footprintLayer = new MapObjectLayer("3D Mapping Footprints", this);
                this.fieldOfViewLayer = new MapObjectLayer("3D Mapping Field of Views", this);
                this.measureLayer = new MapObjectLayer("3D Mapping Measurements", this);
                this.globalLayer = new MapObjectLayer("3D Mapping Globals", this);
                // create the updaters for the layers
                Message.WriteLine("Creating map updaters");
                this.footprintUpdater = new UpdateFootprint(this.footprintLayer, this);
                this.fieldOfViewUpdater = new UpdateFieldOfView(this.fieldOfViewLayer, this);
                this.measurementUpdater = new UpdateMeasurement(this.measureLayer, this);
                this.globalUpdater = new UpdateGlobal(this.measureLayer, this);
                // start the refresh timer
                Message.WriteLine("Creating timer");
                this.refreshTimer = new Timer(250); // milliseconds
                this.refreshTimer.Elapsed += doRefresh;
                this.refreshTimer.AutoReset = true;
                this.refreshTimer.Enabled = true;
                // listen to map draw events to draw our own overlays
                Message.WriteLine("Adding map after-draw listener");
                this.viewX0 = 0.0;
                this.viewY0 = 0.0;
                this.viewX1 = -1.0;
                this.viewX1 = -1.0;
                this.viewBounds = null;
                this.mapExtent = null;
                this.mapScale = 0.0;
                IActiveViewEvents_Event viewEvents = this.map0 as IActiveViewEvents_Event;
                viewEvents.AfterDraw += new IActiveViewEvents_AfterDrawEventHandler(this.drawMapObjects);
                // done
                Message.WriteLine("Created ArcGIS plugin");
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while creating ArcGIS plugin: " + e);
            }
        }

        /// <summary>
        /// Close the plugin
        /// </summary>
        public void Close()
        {
            this.refreshTimer.Stop();
        }

        /// <summary>
        /// Get the map.
        /// </summary>
        /// <returns></returns>
        internal ESRI.ArcGIS.Carto.IMap getMap()
        {
            return this.map0;
        }

        /// <summary>
        /// Get the map extent.
        /// </summary>
        /// <returns></returns>
        internal Bounds getMapExtent()
        {
            return this.mapExtent;
        }

        /// <summary>
        /// Get the map scale.
        /// </summary>
        /// <returns></returns>
        internal double getMapScale()
        {
            return this.mapScale;
        }

        /// <summary>
        /// Get the currently selected feature layer
        /// </summary>
        /// <returns></returns>
        internal IFeatureLayer getSelectedFeatureLayer()
        {
            // get the current vector layer
            IMxDocument mxDocument = (IMxDocument)ArcMap.Application.Document;
            IContentsView contentsView = mxDocument.CurrentContentsView;
            if (contentsView.SelectedItem is IFeatureLayer) return contentsView.SelectedItem as IFeatureLayer;
            return null;
        }

        /// <summary>
        /// get the name of the application
        /// </summary>
        /// <returns>a descriptive name</returns>
        string SDKViewerApplication3DM.GetApplicationName()
        {
            return "ArcMap Plugin by Orbit GT";
        }

        /// <summary>
        /// Check if the plugin is ready for use
        /// </summary>
        /// <returns></returns>
        public bool isReady()
        {
            return this.viewerHost.isViewerReady();
        }

        /// <summary>
        /// Get the 3D Mapping viewer
        /// </summary>
        /// <returns></returns>
        public Viewer3DM getViewer()
        {
            return this.viewerHost.GetViewer();
        }


        /// <summary>
        /// Read the configuration file from disk
        /// </summary>
        /// <returns></returns>
        internal AMap<string> readConfigurationFile()
        {
            AMap<string> configuration = new AMap<string>(Collections.TYPE_STRING);
            if (File.Exists(CONFIG_FILENAME))
            {
                string[] lines = File.ReadAllLines(CONFIG_FILENAME);
                if (lines == null) lines = new string[0];
                foreach (string line in lines)
                {
                    int index = line.IndexOf("=");
                    if (index > 0)
                    {
                        string key = line.Substring(0, index).Trim();
                        string value = line.Substring(index + 1).Trim();
                        if (key.Length > 0) configuration.set(key, value);
                    }
                }
            }
            return configuration;
        }

        /// <summary>
        /// Read the application state file from disk
        /// </summary>
        /// <returns></returns>
        private string readApplicationStateFile()
        {
            string userDir = IOPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
            string fileName = IOPath.Combine(userDir, "arcmap_3dmc_app_state.ini");
            if (File.Exists(fileName) == false) return null;
            string state = File.ReadAllText(fileName);
            if (state != null) Message.WriteLine("Read app state (size " + state.Length + ") from '" + fileName + "'");
            return state;
        }

        /// <summary>
        /// Write the application state file to disk
        /// </summary>
        /// <param name="appState"></param>
        private void writeApplicationStateFile(string appState)
        {
            if (appState == null) return;
            string userDir = IOPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
            if (Directory.Exists(userDir) == false) Directory.CreateDirectory(userDir);
            string fileName = IOPath.Combine(userDir, "arcmap_3dmc_app_state.ini");
            Message.WriteLine("Writing app state (size " + appState.Length + ") to '" + fileName + "'");
            FileStream writer = new FileStream(fileName, FileMode.Create);
            for (int i = 0; i < appState.Length; i++) writer.WriteByte((byte)appState[i]);
            writer.Close();
        }

        /// <summary>
        /// Signal called by the 3D Mapping viewer when the application (login) state has changed. We save the new state to disk to enable auto-login (remember-me) for the next day
        /// </summary>
        /// <param name="value"></param>
        private void OnAppStateChanged(string appState)
        {
            Message.WriteLine("Got 'onAppStateChanged' signal (" + appState + ")");
            this.writeApplicationStateFile(appState);
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer the footprints have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnFootprintsChanged(string message)
        {
            Message.WriteLine("Got 'onFootprintsChanged' signal");
            // request new footprints
            this.footprintUpdater.updatedMapCanvas();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the field-of-views have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnFieldOfViewsChanged(string message)
        {
            Message.WriteLine("Got 'onFieldOfViewsChanged' signal");
            // request new field-of-views
            this.fieldOfViewUpdater.updatedViews();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the measure state has changed
        /// </summary>
        /// <param name="value"></param>
        private void OnMeasureStateChanged(string message)
        {
            Message.WriteLine("Got 'onMeasureStateChanged' signal");
            // request a new measure state
            this.measurementUpdater.updatedMeasurement();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the global map objects have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnGlobalsChanged(string message)
        {
            Message.WriteLine("Got 'onGlobalsChanged' signal");
            // request new globals
            this.globalUpdater.updatedGlobals();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when a zoomout to a spatial extent is requested by the user (zoom-to-extent step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnZoomToExtent(Bounds bounds)
        {
            Message.WriteLine("Got 'onZoomToExtent' signal (" + bounds + ")");
            // Call the update from the UI thread
            this.uiThread.Post(delegate { this.OnZoomToExtent2(bounds); }, null);
        }
        /// <summary>
        /// Do the zoom-to-extent procedure (zoom-to-extent step 2)
        /// </summary>
        /// <param name="bounds">bounds the bounds to zoom the map to</param>
        private void OnZoomToExtent2(Bounds bounds)
        {
            try
            {
                // get the current extent of the map
                IEnvelope mapExtent = this.viewBounds;
                if (mapExtent == null) return;
                ISpatialReference mapCRS = mapExtent.SpatialReference;
                if (mapCRS == null) return;
                IEnvelope newExtent = new EnvelopeClass();
                newExtent.SpatialReference = mapCRS;
                newExtent.PutCoords(mapExtent.XMin, mapExtent.YMin, mapExtent.XMax, mapExtent.YMax);
                Message.WriteLine("Zoom to extent "+bounds+" in map crs "+mapCRS.FactoryCode);
                // calculate the requested extent
                IPoint p0 = MapObjects.createPoint2D(bounds.minX, bounds.minY, MapObjects.VIEWER_CRS);
                p0.Project(mapCRS);
                IPoint p1 = MapObjects.createPoint2D(bounds.maxX, bounds.minY, MapObjects.VIEWER_CRS);
                p1.Project(mapCRS);
                IPoint p2 = MapObjects.createPoint2D(bounds.minX, bounds.maxY, MapObjects.VIEWER_CRS);
                p2.Project(mapCRS);
                IPoint p3 = MapObjects.createPoint2D(bounds.maxX, bounds.maxY, MapObjects.VIEWER_CRS);
                p3.Project(mapCRS);
                double minX = Math.Min(Math.Min(p0.X, p1.X), Math.Min(p2.X, p3.X));
                double minY = Math.Min(Math.Min(p0.Y, p1.Y), Math.Min(p2.Y, p3.Y));
                double maxX = Math.Max(Math.Max(p0.X, p1.X), Math.Max(p2.X, p3.X));
                double maxY = Math.Max(Math.Max(p0.Y, p1.Y), Math.Max(p2.Y, p3.Y));
                // center the extent
                double centerX = 0.5 * (minX + maxX);
                double centerY = 0.5 * (minY + maxY);
                newExtent.CenterAt(MapObjects.createPoint2D(centerX, centerY, mapCRS));
                // scale the extent
                double factorX = (maxX - minX) / (mapExtent.XMax - mapExtent.XMin);
                double factorY = (maxY - minY) / (mapExtent.YMax - mapExtent.YMin);
                double factor = Math.Max(factorX, factorY);
                newExtent.Expand(factor, factor, true/*multiply*/);
                // move the map
                (this.map0 as IActiveView).Extent = newExtent;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while zooming to extent: " + e);
            }
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the viewer needs a location to be picked on the reference map by the user (pick-position step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnPickPositionRequested(string message)
        {
            Message.WriteLine("Got 'onPickPositionRequested' signal");
            // Call the activation from the UI thread
            this.uiThread.Post(delegate { this.OnPickPositionRequested2(); }, null);
        }
        /// <summary>
        /// Activate the pick-position tool (pick-position step 2)
        /// </summary>
        private void OnPickPositionRequested2()
        {
            // activate the tool
            if (ToolPickLocation.INSTANCE!=null) ToolPickLocation.INSTANCE.startPickPosition();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the pick-position operation needs to be cancelled (cancel pick-position step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnPickPositionCancelled(string value)
        {
            Message.WriteLine("Got 'onPickPositionCancelled' signal");
            // deactivate the tool
            if (ToolPickLocation.INSTANCE != null) ToolPickLocation.INSTANCE.stopPickPosition();
        }

        /// <summary>
        /// Call this method to do an update to the map (must be be called from the UI thread).
        /// </summary>
        /// <param name="reason">the reason to redraw the map</param>
        private delegate void DoMapUpdateCallback(string reason);
        internal void DoMapUpdate(string reason)
        {
            Message.WriteLine("Refreshing map");
            //(this.map0 as IActiveView).PartialRefresh(esriViewDrawPhase.esriViewGraphics, null, null); // does not trigger the refresh we expect
            (this.map0 as IActiveView).Refresh();
        }

        /// <summary>
        /// Call this method to request an update to the map (can be called from a non-UI thread).
        /// </summary>
        /// <param name="reason">the reason to redraw the map</param>
        internal void RequestMapUpdate(string reason)
        {
            Message.WriteLine("Requesting map redraw for reason '" + reason + "'");
            if (this.viewerHost == null) return;
            if (this.viewerHost.isViewerReady() == false) return;
            if (this.map0 == null) return;
            // Call the update from the UI thread
            this.uiThread.Post(delegate { this.DoMapUpdate(reason); }, null);
        }

        /// <summary>
        /// Check if the map extent has changed and new map objects need to be fetched
        /// </summary>
        /// <param name="display"></param>
        private void checkChangedMapBounds(IDisplay display)
        {
            // have the screen bounds changed?
            this.viewBounds = display.DisplayTransformation.VisibleBounds;
            if (this.viewBounds == null) return;
            bool movedView = false;
            if (this.viewBounds.XMin != this.viewX0) movedView = true;
            if (this.viewBounds.YMin != this.viewY0) movedView = true;
            if (this.viewBounds.XMax != this.viewX1) movedView = true;
            if (this.viewBounds.YMax != this.viewY1) movedView = true;
            if (movedView == false) return;
            // store the new view bounds
            this.viewX0 = this.viewBounds.XMin;
            this.viewY0 = this.viewBounds.YMin;
            this.viewX1 = this.viewBounds.XMax;
            this.viewY1 = this.viewBounds.YMax;
            Message.WriteLine("moved map");
            this.mapScale = display.DisplayTransformation.ScaleRatio;
            this.mapExtent = MapObjects.transformExtent(this.viewBounds, MapObjects.VIEWER_CRS);
            // request new footprints and field-of-views
            this.footprintUpdater.updatedMapCanvas();
            this.fieldOfViewUpdater.updatedViews();
        }

        /// <summary>
        /// Draw the mapobject overlays on the map (the "AfterDraw" callback from ArcMap after screen refresh)
        /// </summary>
        /// <param name="display">the display to overlay on</param>
        /// <param name="phase">the draw phase</param>
        private void drawMapObjects(IDisplay display, esriViewDrawPhase phase)
        {
            // overlay phase?
            if (phase == esriViewDrawPhase.esriViewGeography || (int)phase == 65535)
            {
                // draw our overlay layers
                this.footprintLayer.Draw(display);
                this.globalLayer.Draw(display);
                this.fieldOfViewLayer.Draw(display);
                this.measureLayer.Draw(display);
                // check if the bounds have changed to request new map objects
                this.checkChangedMapBounds(display);
            }
        }

        // Called by the refresh timer at a regular interval
        //
        static private void doRefresh(Object source, ElapsedEventArgs e)
        {
            PluginArcMap plugin = PluginArcMap.INSTANCE;
            if ((plugin != null) && (plugin.viewerHost.isViewerReady()))
            {
                plugin.footprintUpdater.checkForUpdates();
                plugin.fieldOfViewUpdater.checkForUpdates();
                plugin.measurementUpdater.checkForUpdates();
                plugin.globalUpdater.checkForUpdates();
            }
        }
        
        /// <summary>
        /// Called when the user pressed the 'start' button on the toolbar again
        /// </summary>
        internal void restart()
        {
            // start a new viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
            if (this.viewerHost.hasViewerConnection() == false) this.viewerHost.StartViewer();
            // update all map objects
            this.footprintUpdater.updatedMapCanvas();
            this.fieldOfViewUpdater.updatedViews();
            this.measurementUpdater.updatedMeasurement();
            this.globalUpdater.updatedGlobals();
        }

        /// <summary>
        /// Toolbar action callback method
        /// </summary>
        /// <param name="visible"></param>
        internal void doShowFootprints(bool visible)
        {
            if (this.viewerHost.isViewerReady())
            {
                this.footprintLayer.setVisible(visible);
                this.fieldOfViewLayer.setVisible(visible);
                this.measureLayer.setVisible(visible);
                this.globalLayer.setVisible(visible);
                this.RequestMapUpdate("toggled footprint visibility: " + visible);
            }
        }

        /// <summary>
        /// Toolbar action callback method (called on the UI thread)
        /// </summary>
        internal void doSaveMeasurement()
        {
            try
            {
                Message.WriteLine("doSaveMeasurement");
                if (this.map0 == null) return;
                if (this.viewerHost.isViewerReady())
                {
                    this.measurementUpdater.doSaveMeasurement();
                }
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while saving measurement: " + e);
            }
        }

        /// <summary>
        /// Toolbar action callback method (called on the UI thread)
        /// </summary>
        internal void doAddOverlays()
        {
            try
            {
                Message.WriteLine("doAddOverlays");
                if (this.map0 == null) return;
                if (this.viewerHost.isViewerReady())
                {
                    Overlays.doAddOverlays(this.map0, this.getViewer());
                }
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while adding overlays: " + e);
            }
        }
    }
}
