﻿using System;
using System.Threading;
using System.Threading.Tasks;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class UpdateFootprint manages the updates to the footprint layer
    /// </summary>
    class UpdateFootprint
    {
        private MapObjectLayer layer;
        private PluginArcMap viewerHost;
        private long mapCanvasUpdateTime;
        private long lastRequestTime;
        private bool isUpdating;
        private AList<MapObject> mapObjects;

        /// <summary>
        /// Create a new updater
        /// </summary>
        internal UpdateFootprint(MapObjectLayer layer, PluginArcMap viewerHost)
        {
            this.layer = layer;
            this.viewerHost = viewerHost;
            this.mapCanvasUpdateTime = 0;
            this.lastRequestTime = 0;
            this.isUpdating = false;
            this.mapObjects = null;
        }

        /// <summary>
        /// Called when the mapcanvas has changed and the footprints need to be reloaded
        /// </summary>
        internal void updatedMapCanvas()
        {
            this.mapCanvasUpdateTime = DateTime.Now.Ticks;
        }

        /// <summary>
        /// Check for updates
        /// </summary>
        internal void checkForUpdates()
        {
            //Message.WriteLine("FTPR: check 1");
            // already updating?
            if (this.isUpdating) return;
            //Message.WriteLine("FTPR: check 2");
            // already up-to-date?
            if (this.lastRequestTime > this.mapCanvasUpdateTime) return;
            long currentTime = DateTime.Now.Ticks;
            // throttle the rate of update requests
            if ((this.lastRequestTime > 0) && (currentTime < this.lastRequestTime + 2000 * TimeSpan.TicksPerMillisecond)) return;
            // throttle the rate of mapcanvas changes (fast zoomin operation)
            if ((this.mapCanvasUpdateTime > 0) && (currentTime < this.mapCanvasUpdateTime + 2000 * TimeSpan.TicksPerMillisecond)) return;
            // start the update task (to free the calling timer thread)
            this.isUpdating = true;
            this.lastRequestTime = currentTime;
            Task.Run(() => this.update());
        }

        /// <summary>
        /// Update
        /// </summary>
        private void update()
        {
            Thread.CurrentThread.Name = "update-footprint-task";
            try
            {
                Message.WriteLine("Requesting streaming footprints");
                double mapScale = this.viewerHost.getMapScale();
                if (mapScale <= 0.0) { this.isUpdating = false; return; }
                Bounds mapExtent = this.viewerHost.getMapExtent();
                if (mapExtent == null) { this.isUpdating = false; return; }
                Message.WriteLine("Map extent " + mapExtent + " scale " + mapScale);
                Promise<AList<MapObject>> mapObjectsPromise = this.viewerHost.getViewer().getFootprints2D(mapExtent, mapScale);
                Message.WriteLine("Awaiting footprints with promise id " + mapObjectsPromise.getId());
                this.mapObjects = mapObjectsPromise.waitFor();
                Message.WriteLine("Got " + this.mapObjects.size() + " footprint mapobjects");
                this.isUpdating = false;
                /* Update the layer */
                this.layer.update(this.mapObjects);
                this.viewerHost.RequestMapUpdate("updated footprints");
                return;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while updating footprints: " + e);
                this.isUpdating = false;
                return;
            }
        }
    }
}
