﻿using System;
using System.Threading;
using System.Threading.Tasks;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class UpdateMeasurement manages the updates to the measurement layer
    /// </summary>
    class UpdateMeasurement
    {
        private MapObjectLayer layer;
        private PluginArcMap viewerHost;
        private long updateTime;
        private long lastRequestTime;
        private bool isUpdating;
        private MeasureState measureState;
        private AList<MapObject> mapObjects;

        /// <summary>
        /// Create a new updater
        /// </summary>
        internal UpdateMeasurement(MapObjectLayer layer, PluginArcMap viewerHost)
        {
            this.layer = layer;
            this.viewerHost = viewerHost;
            this.updateTime = 0;
            this.lastRequestTime = 0;
            this.isUpdating = false;
            this.measureState = null;
            this.mapObjects = null;
        }

        /// <summary>
        /// Called when the measurement has been changed and the display need to be updated
        /// </summary>
        internal void updatedMeasurement()
        {
            this.updateTime = DateTime.Now.Ticks;
        }

        /// <summary>
        /// Check for updates
        /// </summary>
        internal void checkForUpdates()
        {
            // already updating?
            if (this.isUpdating) return;
            // already up-to-date?
            if (this.lastRequestTime > this.updateTime) return;
            long currentTime = DateTime.Now.Ticks;
            // throttle the rate of update requests
            if ((this.lastRequestTime > 0) && (currentTime < this.lastRequestTime + 100 * TimeSpan.TicksPerMillisecond)) return;
            // throttle the rate of view changes (fast panning)
            if ((this.updateTime > 0) && (currentTime < this.updateTime + 100 * TimeSpan.TicksPerMillisecond)) return;
            // start the update task (to free the calling timer thread)
            this.isUpdating = true;
            this.lastRequestTime = currentTime;
            Task.Run(() => this.update());
        }

        /// <summary>
        /// Update
        /// </summary>
        private void update()
        {
            Thread.CurrentThread.Name = "update-measurement-task";
            try
            {
                Message.WriteLine("Requesting measurement");
                this.measureState = this.viewerHost.getViewer().getMeasureState();
                if (this.measureState != null) this.mapObjects = this.measureState.mapObjects;
                if (this.mapObjects != null) Message.WriteLine("Got " + this.mapObjects.size() + " mapobjects");
                this.isUpdating = false;
                /* Update the layer */
                this.layer.update(this.mapObjects);
                this.viewerHost.RequestMapUpdate("updated measurement");
                return;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while updating measurement: " + e);
                this.isUpdating = false;
                return;
            }
        }

        // set the value of a named attribute of the new feature object
        //
        private static void setAttribute(IFeature feature,IFields attributeFields, string propertyName,string propertyValue)
        {
            if (propertyName == null) return;
            if (propertyValue == null) return;
            int attributeIndex = attributeFields.FindField(propertyName);
            if (attributeIndex < 0) return;
            IField attributeField = attributeFields.Field[attributeIndex];
            Message.WriteLine("Found attribute '" + attributeField.Name + "' type " + attributeField.Type + " for value '" + propertyValue + "'");
            object attributeValue = null;
            if (attributeField.Type == esriFieldType.esriFieldTypeString) attributeValue = (string)propertyValue;
            else if (attributeField.Type == esriFieldType.esriFieldTypeDouble) attributeValue = Convert.ToDouble(propertyValue);
            else if (attributeField.Type == esriFieldType.esriFieldTypeSingle) attributeValue = Convert.ToSingle(propertyValue);
            else if (attributeField.Type == esriFieldType.esriFieldTypeInteger) attributeValue = (int)Convert.ToDouble(propertyValue);
            else if (attributeField.Type == esriFieldType.esriFieldTypeSmallInteger) attributeValue = (int)Convert.ToDouble(propertyValue);
            else Message.WriteLine("Attribute type not supported");
            if (attributeValue != null)
            {
                feature.set_Value(attributeIndex, attributeValue);
            }
        }

        // Save (add) the current measurement to the active layer (must be called on the UI thread)
        //
        internal void doSaveMeasurement()
        {
            Message.WriteLine("Adding the measurement to the current layer");
            // we need to have a measurement
            if (this.measureState == null)
            {
                Message.WriteLine("There is no current measurement");
                return;
            }
            // print the attributes
            if (this.measureState.properties != null)
            {
                Message.WriteLine("The properties of the measurement are:");
                for (int i = 0; i < this.measureState.properties.size(); i++)
                {
                    MeasureProperty property = this.measureState.properties.get(i);
                    if (property.name != null) Message.WriteLine(">name: '" + property.name + "'");
                    if (property.label != null) Message.WriteLine(">label: '" + property.label + "'");
                    if (property.value != null) Message.WriteLine(">value: '" + property.value + "'");
                    if (property.unit != null) Message.WriteLine(">unit: '" + property.unit + "'");
                }
            }
            // get the current vector layer
            IFeatureLayer featureLayer = this.viewerHost.getSelectedFeatureLayer();
            if (featureLayer == null)
            {
                Message.WriteLine("There is no single selected layer to add to");
                return;
            }
            Message.WriteLine("Found feature layer '"+featureLayer.Name+"'");
            // create the geometry for the layer
            IFeatureClass featureClass = featureLayer.FeatureClass;
            int shapeIndex = featureClass.FindField(featureClass.ShapeFieldName);
            IField shapeField = featureClass.Fields.Field[shapeIndex];
            esriGeometryType layerGeometryType = featureLayer.FeatureClass.ShapeType;
            IGeometry geometry = null;
            // get the (first) mapobject of the measurement
            if (this.measureState.mapObjects.size() < 1)
            {
                Message.WriteLine("The measure state does not have a mapobject");
                return;
            }
            MapObject mapObject = this.measureState.mapObjects.get(0);
            if (layerGeometryType == esriGeometryType.esriGeometryPoint)
            {
                if (mapObject.spatialPoint == null)
                {
                    Message.WriteLine("Cannot add a non-point measurement to a point layer");
                    return;
                }
                Message.WriteLine("Creating a point geometry");
                geometry = MapObjects.getPointGeometry(mapObject, shapeField.GeometryDef.HasZ);
            }
            if (layerGeometryType == esriGeometryType.esriGeometryPolyline)
            {
                if (mapObject.spatialLine == null)
                {
                    Message.WriteLine("Cannot add a non-line measurement to a line layer");
                    return;
                }
                Message.WriteLine("Creating a line geometry");
                geometry = MapObjects.getLineGeometry(mapObject, shapeField.GeometryDef.HasZ, false/*close*/);
            }
            if (layerGeometryType == esriGeometryType.esriGeometryPolygon)
            {
                if (mapObject.spatialArea == null)
                {
                    Message.WriteLine("Cannot add a non-area measurement to an area layer");
                    return;
                }
                Message.WriteLine("Creating an area geometry");
                geometry = MapObjects.getAreaGeometry(mapObject, shapeField.GeometryDef.HasZ);
            }
            if (geometry == null)
            {
                Message.WriteLine("The current layer geometry is neither point, line or polygon but "+ layerGeometryType);
                return;
            }
            // convert the geometry into the CRS of the layer
            ISpatialReference spatialReference = shapeField.GeometryDef.SpatialReference;
            Message.WriteLine("Projecting geometry to feature spatial reference " + spatialReference.FactoryCode);
            geometry.Project(spatialReference);
            // make a new feature for the layer
            Message.WriteLine("Creating new feature");
            Message.WriteLine("Layer has Z? " + shapeField.GeometryDef.HasZ);
            Message.WriteLine("Layer has M? " + shapeField.GeometryDef.HasM);
            IFeature feature = featureClass.CreateFeature();
            Message.WriteLine("Setting geometry");
            feature.Shape = geometry;
            Message.WriteLine("Setting attributes");
            IFields attributeFields = featureClass.Fields;
            // set any default attribute values from the application configuration
            AMap<string> config = this.viewerHost.readConfigurationFile();
            foreach (string key in config.listKeys().toArray())
            {
                if (key.StartsWith("default.attribute.value."))
                {
                    string attributeName = key.Substring(24).Trim();
                    string attributeValue = config.get(key);
                    Message.WriteLine("Applying default attribute '"+ attributeName+"' value '"+attributeValue+"'");
                    setAttribute(feature, attributeFields, attributeName, attributeValue);
                }
            }
            // set attribute values from the measure state
            if (this.measureState.properties != null)
            {
                for (int i = 0; i < this.measureState.properties.size(); i++)
                {
                    MeasureProperty property = this.measureState.properties.get(i);
                    Message.WriteLine("Applying measure property '" + property.name + "' value '" + property.value + "'");
                    setAttribute(feature, attributeFields, property.name, property.value);
                }
            }
            // add the features to the layer
            Message.WriteLine("Adding the new feature to the layer");
            feature.Store();
            Message.WriteLine("Added the new feature to the layer");
            // repaint the map
            this.viewerHost.RequestMapUpdate("added the measurement as a new feature");
        }
    }
}
